/*============================================================================
             DEPTH FIRST VISIT OF A TREE (iterative version)
Grammar for the input tree:  char ::= '0' | '1' | ... | '9'
                             tree ::= char | - char tree | (tree char tree)
Internal tree:               tree = Num  |  Num x tree  |  tree x Num x tree
                                   (Leaf)     (Unary)          (Binary)
Grammar for the output tree: tree ::= n | (n.tree) | (tree.n.tree)
The program uses a stack of trees. Thanks to Nicholas Amadori.
=========================================================================== */
#include <iostream.h>
#include <stdlib.h>   // file needed for the EXIT_SUCCESS and the EXIT_FAILURE
//----------------------------------------------------------------------------
//                    trees
typedef int Values;
enum Tag {leaftree, unarytree, binarytree};
struct tree {
       struct Leaf   {Values value;}; 
       struct Unary  {Values value1; tree* son;};
       struct Binary {Values value2; tree* left; tree* right;};
       union {Leaf l; Unary u; Binary b;};
       Tag tag;};
typedef tree* treeptr;

void printtree(treeptr t)
{switch (t->tag){
case leaftree: cout << t->l.value; break;
case unarytree:
        cout << "(" << t->u.value1 << "."; printtree(t->u.son); cout << ")";
        break;
case binarytree:
        cout << "("; printtree(t->b.left); cout << "." << t->b.value2 << ".";
        printtree(t->b.right); cout << ")";
        break;
default: cout << "Error in printtree!";
        exit(EXIT_FAILURE);}};
//----------------------------------------------------------------------------
//                    from characters to integers
int v(char c){return c-48;};
//----------------------------------------------------------------------------
//                    parsing trees
void parsetree(treeptr &t) { char c;
cin >> c; if ('0' <= c && c <= '9')
            {t->tag=leaftree; t->l.value=v(c);} else
if (c=='-') {t->tag=unarytree; cin >> c; t->u.value1=v(c);
             treeptr t1=new tree; parsetree(t1); t->u.son=t1;} else
if (c=='(') {t->tag=binarytree;
             treeptr t1=new tree; parsetree(t1); t->b.left=t1;
             cin >> c; t->b.value2=v(c);
             treeptr t2=new tree; parsetree(t2); t->b.right=t2;
             cin >> c;}
else {cerr << "Error in parsetree!"; exit(EXIT_FAILURE);}};
//----------------------------------------------------------------------------
//                    stack of trees
class cell {private: treeptr datum; cell* next;
            friend class stack;};
class stack {private: cell* first;
   public:
   stack();
   void push(treeptr x); 
   void pop(treeptr &x);
   bool is_empty();};

stack::stack() {first=NULL;}
void stack::push(treeptr x) {
     cell* newcell=new cell;
     newcell->datum=x; newcell->next=first;
     first=newcell;}
void stack::pop(treeptr &x){
     if (first == NULL) {cerr << "\nError: pop of empty stack!\n";}
     else {x=first->datum; cell* oldfirst=first;
           first=first->next; delete oldfirst;};}
bool stack::is_empty()
     {return first == NULL;}
//----------------------------------------------------------------------------
void main(){
stack st; treeptr pt=new tree;
cout << "Input a tree according to the given grammar above, please.\n";
parsetree(pt);
cout << "\nTree in input:\n"; printtree(pt);
cout << "\n\nDepth first visit of the tree:\n";
st.push(pt);  
while (st.is_empty() == false) {
   treeptr tval=new tree;
   st.pop(pt);
   switch(pt->tag){
   case leaftree:   cout << pt->l.value << " "; break;
   case unarytree:  tval->tag=leaftree;
                    tval->l.value=pt->u.value1;
                    st.push(tval); st.push(pt->u.son); break;
   case binarytree: tval->tag=leaftree;
                    tval->l.value=pt->b.value2;
                    st.push(tval); 
                    st.push(pt->b.right); st.push(pt->b.left); break;
  }  // end switch
};   // end while
cout << endl; 
} 
/* input:    output:
------------------------------------------------------------------------------
             Input a tree according to the given grammar above, please.
             (-231(-564-78))
 
             Tree in input:
             ((2.3).1.((5.6).4.(7.8)))
 
             Depth first visit of the tree:
             3 2 6 5 8 7 4 1     
--------------------------------------------------------------------------- */